@@ -2,34 +2,12 @@ namespace eval wdir {
     global tkWorld
     variable wdir
 
-    # Define the images for the working directory dialog.
-    set wdir(image.refresh) [image create photo -file \
-	    [file join $tkWorld(image_dir) working refresh.gif] \
-	    -format gif]
-    set wdir(image.updir) [image create photo -file \
-	    [file join $tkWorld(image_dir) working updir.gif] \
-	    -format gif]
-    set wdir(image.home) [image create photo -file \
-	    [file join $tkWorld(image_dir) working home.gif] \
-	    -format gif]
-
     # Define the images used by the wdir::contents::dir and
     # wdir::contents::ls procedures.
-    set wdir(image.folder) [image create photo -file \
-	    [file join $tkWorld(image_dir) working folder.gif] \
-	    -format gif]
-    set wdir(image.link) [image create photo -file \
-	    [file join $tkWorld(image_dir) working link.gif] \
-	    -format gif]
-    set wdir(image.exec) [image create photo -file \
-	    [file join $tkWorld(image_dir) working exec.gif] \
-	    -format gif]
-    set wdir(image.socket) [image create photo -file \
-	    [file join $tkWorld(image_dir) working socket.gif] \
-	    -format gif]
-    set wdir(image.file) [image create photo -file \
-	    [file join $tkWorld(image_dir) working file.gif] \
-	    -format gif]
+    foreach f [glob [file join $tkWorld(image_dir) working *.gif]] {
+	set iname [file rootname [file tail $f]]
+	set wdir(image.$iname) [image create photo -file $f -format gif]
+    }
 
     # Define the initial array elements that are not images.
     array set wdir {
@@ -133,40 +111,44 @@ proc wdir::create { } {
     # also lets people type in the starting directory.
     set f1 [frame $wdir(dialog).f1]
     set wdir(entry.wd) [entry $f1.entry_wd \
-	    -width 35 \
-	    -textvariable tkWorld(working_dir)]
-    set wdir(button.home) [button $f1.button_home \
-	    -image $wdir(image.home) \
-	    -command {wdir::change_dir home}]
-    set wdir(button.refresh) [button $f1.button_refresh \
-	    -image $wdir(image.refresh) \
-	    -command wdir::listbox_insert]
-    set wdir(button.up) [button $f1.button_up \
-	    -image $wdir(image.updir) \
-	    -command wdir::up]
-    grid $wdir(entry.wd) $wdir(button.refresh) \
-	    $wdir(button.home) $wdir(button.up) \
-	    -sticky ew \
+	    -textvariable tkWorld(working_dir)]    
+    pack $wdir(entry.wd) \
+	    -fill x \
 	    -padx 2 \
-	    -pady 0 \
+	    -pady 0
 
     # Now create the listbox that we use to display the directories.
     set wdir(listbox.cd) [scroll_listbox::interior \
 	    [scroll_listbox::create $wdir(dialog).lb \
-	    [expr [$wdir(entry.wd) cget -width] + 12] 15]]
+	    [expr [$wdir(entry.wd) cget -width] + 12] 20]]
 
     # Now build the buttons accross the bottom.
     set f2 [frame $wdir(dialog).f2]
-    foreach t "Help Close" {
+    set wdir(button.home) [button $f2.button_home \
+	    -text Home \
+	    -width $wdir(borderwidth) \
+	    -width $wdir(width) \
+	    -height $wdir(height) \
+	    -command {wdir::change_dir home}]
+    foreach t "Refresh Up Help Close" {
 	set b [string tolower $t]
+	switch $b {
+	    "refresh" {
+		set c listbox_insert
+	    }
+	    default {
+		set c $b
+	    }
+	}
 	set wdir(button.$b) [button $f2.button_$b \
 		-text $t \
 		-width $wdir(borderwidth) \
 		-width $wdir(width) \
 		-height $wdir(height) \
-		-command wdir::$b]
+		-command wdir::$c]
     }
-    grid $wdir(button.help) $wdir(button.close) \
+    grid $wdir(button.home) $wdir(button.refresh) $wdir(button.up) \
+	    $wdir(button.help) $wdir(button.close) \
 	    -sticky n \
 	    -padx 2 \
 	    -pady 0
@@ -214,7 +196,7 @@ proc wdir::create { } {
     balloonhelp::for $wdir(button.refresh) \
 	    "Refresh the working directory listing"
     balloonhelp::for $wdir(button.home) \
-	    "Go home"
+	    "Navigate to your home directory"
     balloonhelp::for $wdir(button.up) \
 	    "Chage to the parent directory"
     balloonhelp::for $wdir(listbox.cd) \
@@ -263,25 +245,37 @@ proc wdir::create { } {
 #
 #   None.
 
-proc wdir::change_dir { {p_dir ""} } {
+proc wdir::change_dir { {p_dir PERFORM_SELECTION} } {
     global tkWorld
     variable wdir
 
-    switch $p_dir {
-	home {
+    switch -glob $p_dir {
+	"home" {
 	    set tkWorld(working_dir) $tkWorld(user_home)
+	    wdir::listbox_insert
 	}
-	default {
+	"PERFORM_SELECTION" {
 	    # First get the currect selection index.
 	    set idx [$wdir(listbox.cd) curselection]
 
 	    # Get the directory from the selection index.
 	    set tkWorld(working_dir) [file join \
 		    $tkWorld(working_dir) [$wdir(listbox.cd) get $idx]]
+	    wdir::listbox_insert
+	}
+	".." {
+	    wdir::up
+	}
+	"../*" {
+	    wdir::up
+	    regsub "../" $p_dir {} p_dir
+	    wdir::change_dir $p_dir
+	}
+	default {
+	    regsub "$tkWorld(working_dir)/" $p_dir {} p_dir
+	    set tkWorld(working_dir) [file join $tkWorld(working_dir) $p_dir]
 	}
     }
-
-    wdir::listbox_insert
 }
 
 # wdir::listbox_insert --
@@ -302,7 +296,11 @@ proc wdir::listbox_insert { } {
     variable wdir
 
     # First delete the elements in the listbox
-    $wdir(listbox.cd) delete 0 end
+    if {[string compare {} $wdir(listbox.cd)]} {
+	$wdir(listbox.cd) delete 0 end
+    } else {
+	return
+    }
 
     # List all of the ojects in the current working directory.
     foreach x [lsort [glob -nocomplain \
@@ -428,10 +426,6 @@ proc wdir::tag::enter { t } {
     global tkWorld
 
     # Just change the color scheme.
-#    $tkWorld(log_window) tag configure $t \
-#	    -foreground #ffffff \
-#	    -background #0000aa \
-
     $tkWorld(log_window) tag configure $t \
 	    -foreground #000000 \
 	    -background #d9d9d9
@@ -471,6 +465,7 @@ namespace eval wdir::popup {
 
 proc wdir::popup::create { t x y } {
     global tkWorld
+    variable wdir
 
     # Kill the old command menu if it already exists.
     if [winfo exists .wdir_popup] {
@@ -503,11 +498,14 @@ proc wdir::popup::create { t x y } {
 	    -command [list wdir::popup::rename $t]
 
     switch $ftype {
-	socket -
+	socket {
+	    set img $wdir::wdir(image.socket)
+	}
 	link {
-	    # nothing yet
+	    set img $wdir::wdir(image.link)
 	}
 	directory {
+	    set img $wdir::wdir(image.folder)
 	    $m add command \
 		    -label "Copy" \
 		    -command [list wdir::popup::copy dir $t]
@@ -523,6 +521,7 @@ proc wdir::popup::create { t x y } {
 		    -command [list wdir::popup::filelist ls $t]
 	}
 	file {
+	    set img $wdir::wdir(image.unknown)
 	    $m add command \
 		    -label "Copy" \
 		    -command [list wdir::popup::copy file $t]
@@ -543,22 +542,51 @@ proc wdir::popup::create { t x y } {
 
 	    # Now determine the executable to use if the
 	    # system has it.
-	    switch -glob $ext {
+	    switch -glob [string tolower $ext] {
+		.1 -
+		.2 -
+		.3 -
+		.4 -
+		.5 -
+		.6 -
+		.7 -
+		.8 {
+		    set img $wdir::wdir(image.man)
+		    wdir::popup::filedefault $m $t
+		}
+		.f {
+		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.f_src)
+		}
+		.h {
+		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.h_src)
+		}
+		.cc -
+		.c++ -
+		.c {
+		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.c_src)
+		}
+		
 		.ps {
 		    foreach cmd [system::which "gv ghostview"] {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
 		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.postscript)
 		}
 		.pdf {
 		    foreach cmd [system::which "gv ghostview xpdf"] {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
+		    set img $wdir::wdir(image.pdf)
 		}
 		.fig {
 		    foreach cmd [system::which "xfig"] {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
+		    
 		}
 		.cbb {
 		    foreach cmd [system::which "cbb"] {
@@ -584,18 +612,26 @@ proc wdir::popup::create { t x y } {
 		    foreach cmd [system::which "xv xpaint kpaint"] {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
+		    set img $wdir::wdir(image.image)
 		}
 		.dvi {
 		    foreach cmd [system::which "xdvi gv ghostview"] {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
 		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.dvi)
 		}
 		.tex {
 		    foreach cmd [system::which "tex latex"] {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
 		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.tex)
+		}
+		.prn -
+		.txt {
+		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.txt)
 		}
 		.htm* {
 		    foreach cmd [system::which \
@@ -603,6 +639,14 @@ proc wdir::popup::create { t x y } {
 			wdir::popup::systemadd $m2 $cmd $t
 		    }
 		    wdir::popup::filedefault $m $t
+		    set img $wdir::wdir(image.html)
+		}
+		.java {
+		    set img $wdir::wdir(image.java_src)
+		}
+		.tar -
+		.tgz {
+		    set img $wdir::wdir(image.tgz)
 		}
 		default {
 		    wdir::popup::filedefault $m $t
@@ -615,7 +659,7 @@ proc wdir::popup::create { t x y } {
     #$m add separator
     #$m add command \
 	    -label Properties \
-	    -command [list wdir::popup::properties $t]
+	    -command [list wdir::popup::propbox::create $t $img]
 
     # Create the popup at the mouse point on the screen.
     tk_popup .wdir_popup $x $y
@@ -636,6 +680,11 @@ proc wdir::popup::create { t x y } {
 proc wdir::popup::cc_insert { filename } {
     global tkWorld
 
+    # Put the slash to indicate a directory.
+    if {[file isdirectory [file join $tkWorld(working_dir) $filename]]} {
+	append filename "/"
+    }
+
     $tkWorld(cmd_center) insert end "$filename "
 }
 
@@ -768,7 +817,7 @@ proc wdir::popup::filedefault { m f } {
     $m2 add command \
 	    -label Vi \
 	    -command [list wdir::popup::vi $f]
-    foreach cmd [system::which "kedit xedit axe nedit"] {
+    foreach cmd [system::which "kedit xedit axe jed xjed joe jove nedit"] {
 	wdir::popup::systemadd $m2 $cmd $f
     }
     # Add the preview menu cascade item.
@@ -896,6 +945,93 @@ proc wdir::popup::preview { cmd filename } {
     cmd_center::run
 }
 
+#---------------------------------------------------------------------
+namespace eval wdir::popup::propbox {
+    variable propInfo
+
+    array set propInfo {
+	dialog ""
+	filename.label1 ""
+	filename.label2 ""
+	button.close ""
+    }
+}
+
+# wdir::popup::propbox::create  --
+#
+#   Method to create a properties dialog popup
+#
+# Args
+#
+#   None.
+#
+# Returns
+#
+#   None.
+
+proc wdir::popup::propbox::create { name img } {
+    global tkWorld
+    variable propInfo
+
+    # If the dialog exists, bring it to the front and exit.
+    # Otherwise create the dialog.
+    if [winfo exists $propInfo(dialog)] {
+	switch -- [wm state $propInfo(dialog)] {
+	    normal {
+		raise $propInfo(dialog)
+	    }
+	    withdrawn -
+	    iconic {
+		wm deiconify $propInfo(dialog)
+	    }
+	}
+	focus $propInfo(dialog)
+	return
+    } else {
+	set propInfo(dialog) [toplevel .propInfo \
+		-borderwidth 1 \
+		-relief flat]
+    }
+
+    set f1 [frame $propInfo(dialog).f1 \
+	    -relief groove \
+	    -borderwidth 2]
+    set propInfo(filename.label1) [label $f1.filename_label1 \
+	    -image $img]
+    set propInfo(filename.label2) [label $f1.filename_label2 \
+	    -text $name \
+	    -width 40 \
+	    -anchor w]
+    grid $propInfo(filename.label1) $propInfo(filename.label2) \
+	    -padx 5\
+	    -pady 5 \
+	    -sticky ew
+
+    # Now build the buttons accross the bottom.
+    set f2 [frame $propInfo(dialog).f2]
+    set propInfo(button.close) [button $f2.button_close \
+	    -text Close \
+	    -width 5 \
+	    -command wdir::popup::propbox::close]
+    grid $propInfo(button.close) \
+	    -padx 2 \
+	    -pady 0
+
+    pack $f1 $f2 \
+	    -side top \
+	    -padx 5 \
+	    -pady 5 \
+	    -ipadx 5 \
+	    -ipady 5
+}
+
+proc wdir::popup::propbox::close { } {
+    variable propInfo
+
+    balloonhelp::cancel
+    destroy $propInfo(dialog)
+}    
+
 #---------------------------------------------------------------------
 namespace eval wdir::contents {
     variable contentInfo
@@ -966,6 +1102,7 @@ proc wdir::contents::header { } {
 	    -anchor w \
 	    -foreground #ffffff \
 	    -background #00009f \
+	    -relief raised \
 	    -width [$tkWorld(log_window) cget -width]]
 
     # Put the label in the LW.
@@ -999,7 +1136,8 @@ proc wdir::contents::dir { } {
 
     # First get the list of all the file in the current
     # working directory.
-    set x(list) [lsort [glob -nocomplain $tkWorld(working_dir)/*]]
+    set x(list) [lsort [glob -nocomplain $tkWorld(working_dir)/.* \
+	    $tkWorld(working_dir)/*]]
 
     if {![string length $x(list)]} {
 	return
@@ -1007,7 +1145,6 @@ proc wdir::contents::dir { } {
 
     # Build the listing of all of the directories and find the
     # length of the longest directory name.
-    set x(dirlist) ""
     set maxlen 0
     foreach f $x(list) {
 	set len [expr [string length [file tail $f]] + 2]
@@ -1071,16 +1208,30 @@ proc wdir::contents::dir { } {
 
 proc wdir::contents::ls { } {
     global tkWorld
+    variable wdir
 
     $tkWorld(log_window) insert end "\n"
 
     # First get the list of all the file in the current
     # working directory.
-    set x(list) [lsort [glob -nocomplain $tkWorld(working_dir)/*]]
+    set x(list) [lsort [glob -nocomplain $tkWorld(working_dir)/.* \
+	    $tkWorld(working_dir)/*]]
 
     foreach f $x(list) {
-	# First get the file type and define the image to display
-	set x(ftype) [file type $f]
+	# First get the file type and define the image to display.
+	# If it is a link, the resolve the link.
+	#set x(ftype) [file type $f]
+	if {[set x(ftype) [file type $f]] == "link"} {
+	    if [catch {[wdir::resolve_link $f]} errmsg] {
+		$tkWorld(log_window) image create end \
+			-image $wdir::wdir(image.smallwarning) \
+			-padx 2 \
+			-pady 0
+		$tkWorld(log_window) insert end \
+			"Warning: $errmsg\n"
+		continue
+	    }
+	}
 	set img [wdir::parse::filetype $f]
 
 	# Get the permissions for the file
@@ -1095,9 +1246,13 @@ proc wdir::contents::ls { } {
 	set x(fsize) [file size $f]
 
 	# Get the number of link and the file create time
-	if {[catch {file stat $f y}]} {
+	if {[catch {file stat $f y} errmsg]} {
+	    $tkWorld(log_window) image create end \
+		    -image $wdir::wdir(image.smallwarning) \
+		    -padx 2 \
+		    -pady 0
 	    $tkWorld(log_window) insert end \
-		    "couldn't stat file: [file tail $f]\n"
+		    "Warning: $errmsg\n"
 	    continue
 	}
 	set x(nlink) $y(nlink)
@@ -1141,16 +1296,15 @@ namespace eval wdir::parse {
 # wdir::parse::filetype  --
 #
 #   Method to decipher the prefix string or suffix string
-#   based on the file type.
+#   based on the file type and return the image to use.
 #
 # Args
 #
-#   t    - The file type i.e. file, directory, etc.
-#   rule - Either prefix or suffix.
+#   f - The file to determine the type to.
 #
 # Returns
 #
-#   The prefix or suffix string.
+#   The image identifier for the image to use.
 
 proc wdir::parse::filetype { f } {
     variable wdir
@@ -1158,20 +1312,20 @@ proc wdir::parse::filetype { f } {
     # Based on the file type, return the image
     switch [file type $f] {
 	directory {
-	    set img $wdir::wdir(image.folder)
+	    set img $wdir::wdir(image.smallfolder)
 	}
 	link {
-	    set img $wdir::wdir(image.link)
+	    set img $wdir::wdir(image.smalllink)
 	}
 	socket {
-	    set img $wdir::wdir(image.socket)
+	    set img $wdir::wdir(image.smallsocket)
 	}
 	default {
-	    set img $wdir::wdir(image.file)
+	    set img $wdir::wdir(image.smallfile)
 	}
     }
     if {[file type $f] == "file" && [file executable $f]} {
-	set img $wdir::wdir(image.exec)
+	set img $wdir::wdir(image.smallexec)
     }
     
     return $img
@@ -1251,3 +1405,181 @@ proc wdir::parse::permissions { p } {
     # Send back the full string i.e. rwxr-xr-x.
     return $perm_str
 }
+
+#---------------------------------------------------------------------
+namespace eval wdir::expand {
+    # No private variables
+}
+
+# wdir::expand::filename  --
+#
+#   Method to perform filename completion 
+#
+# Args
+#
+#   None.
+#
+# Returns
+#
+#   None.
+
+proc wdir::expand::filename { } {
+    global tkWorld
+    variable wdir
+
+    set t $tkWorld(cmd_center)
+
+    # Define the indices used to find the string to match.
+    # Start from the cursor and work backward to find the text
+    # to match.  Assume a string in the CC of the form:
+    #
+    # cmd xxx<tab>
+    #
+    set iend [$t index insert]
+    set istart [$t search -backward -exact " " insert 0.0]
+
+    # If the user has not entered a command, then disable filename
+    # completion in the command center.
+    if {![string compare {} $istart]} {
+	stderr::log 0006
+	return
+    }
+
+    # Get the string to match.
+    set m [$t get "$istart +1c" $iend]
+    
+    # If a command is entered, but the user has not enter a partial
+    # string, then disable filename completion.
+    if {![string compare {} $m]} {
+	bell
+	return
+    }
+    
+    # Define the list to search from.  If there is no match, then
+    # let the user know and do not complete the filename.
+    if {[catch {glob [file join $tkWorld(working_dir) $m*]} flist]} {
+	bell
+	$t delete $iend end
+	return
+    } else {
+	# Retrieve the best match from the users partial string.
+	set match [wdir::expand::bestmatch $flist]
+
+	# Process if a match is found.
+	if {[string compare {} $match]} {
+	    # Remove the extra characters from the CC.
+	    $t delete "$istart +1c" $iend
+
+	    # Create the insert string with the working directory
+	    # stripped off and then insert it.
+	    regsub "$tkWorld(working_dir)/" $match {} istr
+	    $t insert end $istr
+
+	    # Append the unix directory indicator character or
+	    # if necessary, popup a list of possible completions.
+	    if {[file isdirectory $match] && \
+		    [string range $match end end] != "/"} {
+		$t insert end "/"
+	    } elseif {[llength [set mlist [glob $match*]]] > 1} {
+		# Strip out the working directory for each element
+		# in the expansion list and in the match.
+		regsub -all "$tkWorld(working_dir)/" $mlist {} mlist
+		regsub "$tkWorld(working_dir)/" $match {} match
+
+		# Now create the popup with the matches.
+		wdir::expand::selection $mlist $match
+	    }
+	}
+    }
+}
+
+# wdir::expand::bestmatch  --
+#
+#   This procedure taken from Jeff Hobbs tkcon.tcl package.
+#
+#   tkConExpandBestMatch - finds the best unique match in a list of
+#   names.  The extra $e in this argument allows us to limit the
+#   innermost loop a little further.  This improves speed as $l
+#   becomes large or $e becomes long.
+#
+# Args
+#
+#   l - list to find best unique match in.
+#   e - currently best known unique match.
+#
+# Returns
+#
+#   The longest unique match in the list.
+
+proc wdir::expand::bestmatch {l {e {}}} {
+    set ec [lindex $l 0]
+    if {[llength $l]>1} {
+	set e  [string length $e]; incr e -1
+	set ei [string length $ec]; incr ei -1
+	foreach l $l {
+	    while {$ei>=$e && [string first $ec $l]} {
+		set ec [string range $ec 0 [incr ei -1]]
+	    }
+	}
+    }
+    return $ec
+}
+
+# wdir::expand::selection  --
+#
+#   Method to create a popup dialog containing all of the possible
+#   filename completions.
+#
+# Args
+#
+#   flist - The list of filename completions.
+#   match - The partial match that the user has keyed in with
+#           the tkWorld working directory embedded in it.
+#
+# Returns
+#
+#   None.
+
+proc wdir::expand::selection { flist match } {
+    global tkWorld
+
+    # Build the filename completion dialog.
+    if {[winfo exists .wdir_filename_selection]} {
+	destroy .wdir_filename_selection
+    }
+
+    # Define the parent menu to place the completion list in.
+    set m [menu .wdir_filename_selection\
+	    -tearoff 0]
+
+    # Add each file to the menu for a completion.
+    foreach f $flist {
+	# Set the string to insert into the CC based on the remaining
+	# characters that make the completion: World completes tkWorld
+	# if the user typed in tk.
+	regsub $match $f {} istr
+
+	# Insert the typical unix directory indicator to the insert
+	# string.
+	if {[file isdirectory [file join $tkWorld(working_dir) $f]]} {
+	    append istr "/"
+	}
+	
+	# Add the menu item which inserts the insert string into the CC.
+	$m add command \
+		-label $f \
+		-command "$tkWorld(cmd_center) insert insert $istr"
+    }
+
+    # Now define where to put the popup based on the CC position.
+    set x [winfo rootx $tkWorld(cmd_center)]
+    set y [expr [winfo rooty $tkWorld(cmd_center)] + 2]
+
+    # For the user, push the popup to the right based on the current
+    # location of the insert cursor.
+    incr x [lindex [$tkWorld(cmd_center) bbox insert] 0]
+    incr y [lindex [$tkWorld(cmd_center) bbox insert] 1]
+
+    # Create the menu to select completions from.
+    tk_popup .wdir_filename_selection $x $y
+}
\ No newline at end of file
